#include "config.h"
#define _XOPEN_SOURCE
#define __EXTENSIONS__
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <signal.h>
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif

#include <fstream.h>

#include "tcp.h"

#ifndef O_NDELAY
#ifdef O_NONBLOCK
#define O_NDELAY O_NONBLOCK
#else
#error neither O_NDELAY nor O_NONBLOCK is defined!  Please fix the includes.
#endif
#endif

#ifdef _LINUX_C_LIB_VERSION_MAJOR
#define HAVE_SYS_ERRLIST
#endif
#ifndef HAVE_STRERROR
#ifndef HAVE_SYS_ERRLIST
extern char** sys_errlist;
#endif
#endif

ostream& operator<<(ostream &o,const NetworkError &e) {
#ifdef HAVE_STRERROR
  o << e.message << ": " << strerror(e.Errno);
#else
  o << e.message << ": " << sys_errlist[e.Errno];
#endif
  return o;
}

/* Open a server socket, return a file descriptor ready for accept() or
 * select() or -1 on error.  The socket is set to non-blocking I/O and
 * address reuse. */
int selectable_server_socket_magic(short port,long addr) {
/* e.g.: port=80, addr=INADDR_ANY */
  static int True=1;	/* needed for setsockopt :( */
  int s;
  struct sockaddr_in sa;
  if ((s=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP))==-1)
    throw NetworkError("socket",errno);
//  if (fcntl(s,F_SETFL,O_NDELAY)==-1) {
//    perror("fcntl");
//    return -1;
//  }
  if (setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(void *)&True, sizeof(True))==-1)
    throw NetworkError("setsockopt(SO_REUSEADDR)",errno);
  sa.sin_family=AF_INET;
  sa.sin_port=htons(port);
  sa.sin_addr.s_addr=htonl(addr);
  if (bind(s,(struct sockaddr *)&sa,sizeof(sa))==-1)
    throw NetworkError("bind",errno);
  if (listen(s,100)==-1)
    throw NetworkError("listen",errno);
  return s;
}

/* Open a server socket, return a file descriptor ready for accept() or
 * select() or -1 on error.  The socket is set to non-blocking I/O and
 * address reuse. */
int server_socket_magic(short port,long addr) {
/* e.g.: port=80, addr=INADDR_ANY */
  int s=selectable_server_socket_magic(port,addr);
  struct sockaddr_in sa;
  int addrlen=sizeof(sa);
  if ((s=accept(s,(struct sockaddr *)&sa,&addrlen))==-1)
    throw NetworkError("accept",errno);
  return s;
}

/* Trivial DNS lookup routine that accepts a string that can contain a
 * host name or IP address and returns 0 or the host address in network
 * order. */
unsigned long DNSLookup(const char *Host) {
  struct hostent *he;
  if ((he = gethostbyname(Host)) == NULL)
    if ((he = gethostbyaddr(Host, strlen(Host), AF_INET)) == NULL)
      throw DNSError();
  return *(long*)(he->h_addr_list[0]);
}

/* Open a client socket to the specified host and port (in host order).
 * Returns a file descriptor or -1 on error. */
int client_socket_magic(const char *host,short port) {
  unsigned long target_address;
  struct sockaddr_in inp;
  int s;
  target_address=DNSLookup(host);
  if ((s=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP))==-1)
    throw NetworkError("socket",errno);
  inp.sin_family = AF_INET;
  inp.sin_addr.s_addr=target_address;
  inp.sin_port = htons(port);
  if (connect(s,(struct sockaddr *)&inp,sizeof(inp)) == -1)
    throw NetworkError("connect",errno);
  return s;
}

tcpstream::tcpstream(const char *hostname,short port) :
  fstream(client_socket_magic(hostname,port)) { }

tcpstream::tcpstream(short port) :
  fstream(server_socket_magic(port,INADDR_ANY)) { }

int pipe_magic(const char *program,char *const commandline[],pid_t &PID,
	       bool read=true,bool write=true) {
  int sockets[2];
  if (socketpair(AF_UNIX,SOCK_STREAM,PF_UNSPEC,sockets)) {
    throw NetworkError("socketpair",errno);
  }
  if (!(PID=fork())) {
    close(sockets[1]);
    if (write)
      dup2(sockets[0],0);
    if (read)
      dup2(sockets[0],1);
    execvp(program,commandline);
    throw NetworkError("execv",errno);
  }
  close(sockets[0]);
  return sockets[1];
}

pipestream::pipestream(const char *program,char *const commandline[]) :
  fstream(pipe_magic(program,commandline,PID)) { }

pipestream::~pipestream() {
  kill(PID,15);
  waitpid(PID,0,0);
}

